package se.lth.dit.odinmp;

import java.util.*;
import java.io.*;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.syntaxtree.*;
import se.lth.dit.odinmp.visitor.*;


public class OdinMP {
    // Variables that keep track of things we need in parsing
    public Vector sourceFiles = new Vector();
    public SourceFile currentSourceFile;
    public Vector globalTypedefs;
    public SymbolTable threadprivateSymbols = new SymbolTable();
    public String generated;
    public static String cppOptions = new String();
    public static String includes = null;

    // constants
    public static final String OmpAttrShared = "shared";
    public static final String OmpAttrPrivate = "private";
    public static final String OmpAttrFirstprivate = "firstprivate";
    public static final String OmpAttrLastprivate = "lastprivate";
    public static final String OmpAttrThreadprivate = "threadprivate";
    public static final String OmpAttrReduction = "reduction";
    public static final String OmpAttrNowait = "nowait";
    public static final String OmpAttrOrdered = "ordered";
    
    public static final int mode_Never = 0;
    public static final int mode_Runtime = 1;
    public static final int mode_Always = 2;

    public static Properties defaults = new Properties(System.getProperties());
    public static Properties properties = new Properties(defaults);
    public static Properties arguments = new Properties(properties);

    public static Vector sourceFileNames = new Vector();

    public static boolean debugMode = false;
    
    // identity string
    static public String ident = "OdinMP/C->Cp";
    static public String longIdent = ident + " (C -> C with POSIX threads) v1.0";

    public static boolean debugMode() {
	return new Boolean(arguments.getProperty("odinmp.debug", Boolean.FALSE.toString())).booleanValue();
    }

    public static int targetDebugMode() {
	return new Integer(arguments.getProperty("odinmp.targetDebugMode", new Integer(mode_Never).toString())).intValue();
    }

    public static int targetTimingMode () {
	return new Integer(arguments.getProperty("odinmp.targetTimingMode", new Integer(mode_Never).toString())).intValue();
    }

    public static String cpp () {
	return arguments.getProperty("odinmp.cpp", "cppp") + " " + arguments.getProperty("odinmp.cflags", "");
    }

    public static String cc () {
	return arguments.getProperty("odinmp.cc", "gcc") + " " + arguments.getProperty("odinmp.ldflags", "");
    }

    public static String numProcessors () {
	return arguments.getProperty("odinmp.numProcessors", "4");
    }

    public static String numThreads () {
	return arguments.getProperty("odinmp.numThreads", "4");
    }

    public static String targetDirectory () {
	return arguments.getProperty("odinmp.targetDirectory", "odinmp_ccd");
    }

    public static String supporting () {
	return arguments.getProperty("odinmp.supporting", "supporting");
    }

    public OdinMP () {
    }

    public SourceFile addFile (File f, String realFilename) {
	currentSourceFile = new SourceFile(this, f, realFilename);
	sourceFiles.addElement(currentSourceFile);
	return currentSourceFile;
    }

    public static void initProperties() {
	String odinmp_home = System.getProperty("odinmp.home");
	String odinmp_args = System.getProperty("odinmp.arguments");
	File f;
	try {
	    f = new File(odinmp_home, "odinmp.defaults");
	    if (f.canRead()) 
		defaults.load(new FileInputStream(f));
	} catch (Exception ex) {
	    System.err.println("Failed to load defaults: ``" + ex + "''");
	}

	try {
	    f = new File("odinmp.properties");
	    if (f.canRead()) 
		properties.load(new FileInputStream(f));
	} catch (Exception ex) {
	    System.err.println("Failed to load properties: ``" + ex + "''");
	}
	
	try {
	    f = new File(odinmp_args);
	    debugPrintln("arguments file = ``" + odinmp_args + "''");
	    if (f.canRead()) 
		arguments.load(new FileInputStream(f));
	} catch (Exception ex) {
	    System.err.println("Failed to load arguments: ``" + ex + "''");
	}
    }

    public static int modeForString(String arg, String modeName) {
	if (arg.startsWith("y") ||
	    arg.startsWith("Y") ||
	    arg.startsWith("a") ||
	    arg.startsWith("A")) {
	    return mode_Always;
	} else if (arg.startsWith("n") ||
		   arg.startsWith("N")) {
	    return mode_Never;
	} else if (arg.startsWith("r") ||
		   arg.startsWith("R")) {
	    return mode_Runtime;
	} else {
	    System.err.println("unknown " + modeName + " mode: ``" + arg + "''");
	    System.err.println("  use `never', `runtime' or `always'");
	    return -1;
	}
    }

    public static void parseArguments(String[] args) {
	int i;

	for (i = 0; i < args.length; i++) {
	    if (args[i].equals("--debug")) {
		debugMode = true;
		continue;
	    }

	    if (args[i].startsWith("--targetDebug=")) {
		String arg = args[i].substring(args[i].indexOf("=")+1);
		int mode = modeForString(arg, "target debug");
		if (mode >= 0) {
		    arguments.put("odinmp.targetDebugMode", new Integer(mode).toString());
		}
		/* debugPrintln("odinmp.targetDebugMode = `" + mode + "'"); */
		continue;
	    }

	    if (args[i].startsWith("--targetTiming=")) {
		String arg = args[i].substring(args[i].indexOf("=")+1);
		int mode = modeForString(arg, "target timing");
		if (mode >= 0) {
		    arguments.put("odinmp.targetTimingMode", new Integer(mode).toString());
		}
		/* debugPrintln("odinmp.targetTimingMode = `" + mode + "'"); */
		continue;
	    }
	
	    if (args[i].startsWith("--cpp=")) {
		String arg = args[i].substring(args[i].indexOf("=")+1);
		arguments.put("odinmp.cpp", arg);
		/* debugPrintln("odinmp.cpp = `" + arg + "'"); */
		continue;
	    }

	    if (args[i].startsWith("--cc=")) {
		String arg = args[i].substring(args[i].indexOf("=")+1);
		arguments.put("odinmp.cc", arg);
		/* debugPrintln("odinmp.cc = `" + arg + "'"); */
		continue;
	    }

	    if (args[i].startsWith("--targetDirectory=")) {
		String arg = args[i].substring(args[i].indexOf("=")+1);
		arguments.put("odinmp.targetDirectory", arg);
		/* debugPrintln("odinmp.targetDirectory = `" + arg + "'"); */
		continue;
	    }

	    if (args[i].startsWith("--numProcessors=")) {
		String arg = args[i].substring(args[i].indexOf("=")+1);
		arguments.put("odinmp.numProcessors", arg);
		/* debugPrintln("odinmp.numProcessors = `" + arg + "'"); */
		continue;
	    }

	    if (args[i].startsWith("--numThreads=")) {
		String arg = args[i].substring(args[i].indexOf("=")+1);
		arguments.put("odinmp.numThreads", arg);
		/* debugPrintln("odinmp.numThreads = `" + arg + "'"); */
		continue;
	    }

	    if (args[i].startsWith("-")) {
		cppOptions = cppOptions + args[i];
	    }

	    sourceFileNames.addElement(args[i]);
	}

	if (debugMode) {
	    Enumeration e = arguments.propertyNames();
	    while (e.hasMoreElements()) {
		String key = (String)e.nextElement();
		System.err.println("``" + key + "'' => ``" + arguments.getProperty(key) + "''");
	    }
	}
    }

    // Run the parser
    public static void main (String args[]) {
	OdinMP odinmp = new OdinMP();
	TranslationUnit root = null;
	Enumeration e;

	System.err.println(longIdent);

	initProperties();
	parseArguments(args);

	e = sourceFileNames.elements();
	while (e.hasMoreElements()) {
	    String s = (String)e.nextElement();
	    odinmp.addFile(new File(s), s);
	}

	e = odinmp.sourceFiles.elements();
	while (e.hasMoreElements()) {
	    SourceFile sf = odinmp.currentSourceFile = (SourceFile)e.nextElement();
	    
	    System.err.print(ident + ":  File ``" + sf.name + "'': parsing ... ");

	    root = sf.parse();

	    if (root == null) {
		System.err.println("failed!!!");
		continue;
	    }
	    System.err.println("ok.");

	    DataBuildingVisitor dbv = new DataBuildingVisitor(odinmp);
	    sf.dbv = dbv;

	    System.err.print(OdinMP.ident + ":  File ``" + sf.name + "'': building internal data structures ... ");	    
	    root.accept(dbv);
	    System.err.println("ok.");
	}

	System.err.println(ident + ": All source files internalized, proceeding\n");
	
	e = odinmp.sourceFiles.elements();
	while (e.hasMoreElements()) {
	    SourceFile sf = (SourceFile)e.nextElement();
	    odinmp.currentSourceFile = sf;
	    DataBuildingVisitor dbv = sf.dbv;
	    FileOutputStream fos;

	    System.err.print(ident + ": File ``" +sf.name+"'': unparsing ... ");

	    try {
		fos = new FileOutputStream(new File(targetDirectory(), sf.name));
	    } catch (Exception ex) {
		System.err.println("\nCaught exception: " + e);
		continue;
	    }

	    PrintWriter pw = new PrintWriter(fos);

	    pw.println("/* Typedefs: */");
	    dbv.printTypedefsTo(pw);
	    pw.println("");
	    pw.println("/* Common stuff for all parsed files follows! */");
	    pw.println("/* Generated private & shared memory data types: */");
	    pw.println(odinmp.commonTypedefs(dbv.typedefs));
	    pw.println(odinmp.PRINTF_define());
	    pw.println(odinmp.TIMING_define());
	    pw.println("/* External declarations for OdinMP: */");
	    pw.println(odinmp.odinmpExternals());
	    pw.println("/* Generated Token String: */");
	    dbv.printCodeTo(pw);
	    pw.println("/* Generated extra code: */");
	    dbv.printExtraCodeTo(pw);

	    pw.close(); pw = null;
	    try {
		fos.close(); 
	    } catch (Exception ex) {
		System.err.println("Caught exception: " + e);
		continue;
	    }
	    fos = null;

	    System.err.println("ok.");
	}

	System.err.print(ident + ": generating odinmp.c ... ");
	FileOutputStream fos = null;
	
	try {
	    fos = new FileOutputStream(new File(targetDirectory(), "odinmp.c"));
	} catch (Exception ex) {
	    System.err.println("Caught exception: " + e);
	}
	
	PrintWriter pw = new PrintWriter(fos);
	pw.println(odinmp.odinmpRoutines());

	pw.close(); pw = null;
	try {
	    fos.close(); 
	} catch (Exception ex) {
	    System.err.println("Caught exception: " + e);
	}
	fos = null;

	System.err.println("ok.");
    }   

    public String threadprivateDataTypedef() {
	if (threadprivateSymbols.symbolsInOrder.size() == 0) {
	    return "";
	}

	StringBuffer retval = new StringBuffer();
	retval.append("typedef struct {\n");
	Enumeration e = threadprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    Symbol sym = (Symbol)e.nextElement();
	    retval.append("  " + sym.typeString(" ( * " + sym.name + " ) ") + ";\n");
	}
	retval.append("} odinmp_threadprivate_t;\n\n");
	return new String(retval);
    }
    
    public String threadprivateData() {
	if (threadprivateSymbols.symbolsInOrder.size() == 0) {
	    return "";
	}

	return "  odinmp_threadprivate_t threadprivate;\n";
    }

    public Vector threadprivateDependsOnTypes() {
	Vector result = new Vector();
	Enumeration e = threadprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    addToVector(result, ((Symbol)e.nextElement()).dependsOnTypes());
	}
	return result;
    }

    public static void addToVector(Vector target, Vector source) {
	Enumeration e;
	for (e = source.elements(); e.hasMoreElements();) {
	    Object o = e.nextElement();
	    if (!target.contains(o))
		target.addElement(o);
	}
    }

    public String commonTypedefs(SymbolTable existingTypedefs) {
	Enumeration e;

	if (generated != null) {
	    StringBuffer retval = new StringBuffer();
	    /* debugPrintln("Extra needed types:"); */
	    e = globalTypedefs.elements();
	    while (e.hasMoreElements()) {
		Symbol type = (Symbol)e.nextElement();
		if (existingTypedefs.symbolWithName(type.name) == null) {
		    /* debugPrintln(type.toString()); */
		    retval.append(type.toString() + ";\n");
		}
	    }

	    retval.append(generated);
	    return new String(retval);
	}

	int n;
	IncrementalReader ir;

	try {
	    ir = new IncrementalReader(new BufferedReader(new FileReader(new File(supporting(), "odinmp_typedefs.h"))));
	} catch (Exception ex) {
	    System.err.println("Caught Exception trying to open incremental file: ``" + ex + "''");
	    return null;
	}

	globalTypedefs = new Vector();

	generated = 
	    "#define ODINMP_NUM_THREADS " + numThreads() + "\n" +
	    "#define ODINMP_NUM_PROCESSORS " + numProcessors() + "\n\n";

	generated = generated + 
	    ir.readIncrement();

	if (targetTimingMode() != mode_Never)
	generated = generated +
	    "  double timer_wait;\n";

	generated = generated + 
	    ir.readIncrement();

	String td_prmdefs = "";
	String td_prm = "/* typedef */ struct _odinmp_thread_data_t {\nodinmp_thread_info_t thread_info;\n";
	String td_shmdefs = "";
	String td_shm = "/* typedef */ struct _odinmp_shared_data_t {\n  int dummy;\n";

	String lap_infos = "typedef struct {\n  int dummy;\n";
	String tmp;

	/* parallel regions */
	e = ParallelConstructDescriptor.inOrder.elements();
	while (e.hasMoreElements()) {
	    ParallelConstructDescriptor pcd = (ParallelConstructDescriptor)e.nextElement();
	    td_prmdefs = td_prmdefs + pcd.threadDataTypedef();
	    td_prm = td_prm + pcd.threadData();
	    
	    td_shmdefs = td_shmdefs + pcd.sharedDataTypedef();
	    td_shm = td_shm + pcd.sharedData();

	    addToVector(globalTypedefs, pcd.dependsOnTypes());
	}

	/* for regions */
	e = ForConstructDescriptor.inOrder.elements();
	while (e.hasMoreElements()) {
	    ForConstructDescriptor fcd = (ForConstructDescriptor)e.nextElement();
	    lap_infos = lap_infos + "  odinmp_for_lap_info_t " + fcd + ";\n";
	}
	

	/* sections */
	e = SectionsConstructDescriptor.inOrder.elements();
	while (e.hasMoreElements()) {
	    SectionsConstructDescriptor scd = (SectionsConstructDescriptor)e.nextElement();
	    lap_infos = lap_infos + "  odinmp_sect_lap_info_t " + scd + ";\n";
	}

	/* singles */
	e = SingleConstructDescriptor.inOrder.elements();
	while (e.hasMoreElements()) {
	    SingleConstructDescriptor scd = (SingleConstructDescriptor)e.nextElement();
	    lap_infos = lap_infos + "  odinmp_sing_lap_info_t " + scd + ";\n";
	}

	/* threadprivate */
	td_prmdefs = td_prmdefs + threadprivateDataTypedef();
	td_prm = td_prm + threadprivateData();
	addToVector(globalTypedefs, threadprivateDependsOnTypes());


	td_prm = td_prm + "};\n";
	td_shm = td_shm + "};\n";
	lap_infos = lap_infos + "} odinmp_thread_lap_info_t;\n";

	generated = generated +
	    td_prmdefs + "\n" + 
	    td_prm + "\n" + 
	    td_shmdefs + "\n" + 
	    td_shm + "\n" +
	    lap_infos + "\n";

	generated = generated +
	    ir.readIncrement();
	
	if ((n = ForConstructDescriptor.nextIndex) > 0) {
	    generated = generated + 
		"  odinmp_for_lap_list_t *for_laps[" + n + "];\n" +
		"  odinmp_barrier_t for_barriers[" + n + "];\n" +
		"\n";
	}

	if ((n = SectionsConstructDescriptor.nextIndex) > 0) {
	    generated = generated + 
		"  odinmp_sect_lap_list_t *sect_laps[" + n + "];\n" +
		"  odinmp_barrier_t sect_barriers[" + n + "];\n" +
		"\n";
	}

	if ((n = SingleConstructDescriptor.nextIndex) > 0) {
	    generated = generated + 
		"  int sing_laps[" + n + "];\n" +
		"  odinmp_barrier_t sing_barriers[" + n + "];\n" +
		"\n";
	}

	if ((n = BarrierConstructDescriptor.nextIndex) > 0) {
	    generated = generated + 
		"  odinmp_barrier_t barriers[" + n + "];\n" +
		"\n";
	}
	
	generated = generated +
	    ir.readIncrement();

	try {
	    ir.close();
	} catch (Exception ex) {
	    System.err.println("Caught Exception trying to close incremental file: ``" + ex + "''");
	}

	return commonTypedefs(existingTypedefs);
    }

    public String odinmpVariables() {
	StringBuffer retval = new StringBuffer();
	Enumeration e;

	// locks for critical sections
	return new String(retval);
    }

    public String reportTimingWsc(String tag, Enumeration e) {
	StringBuffer retval = new StringBuffer();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    String s = o.toString();
	    
	    retval.append("   fprintf(stderr,\"" + tag + " " + s + ":\\n\");\n");
	    retval.append("   fprintf(stderr,\"  executed %d times total\\n\", odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("   if (odinmp_tacc_" + s + "_total.n > 0) {\n");
	    retval.append("      fprintf(stderr,\"  total time:         min = %#12.7g s\\n\", odinmp_tacc_" + s + "_total.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per thread\\n\", odinmp_tacc_" + s + "_total.total / odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s\\n\", odinmp_tacc_" + s + "_total.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s\\n\", odinmp_tacc_" + s + "_total.total);\n");
	    retval.append("      fprintf(stderr,\"  working time:       min = %#12.7g s\\n\", odinmp_tacc_" + s + "_work.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per thread\\n\", odinmp_tacc_" + s + "_work.total / odinmp_tacc_" + s + "_work.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s\\n\", odinmp_tacc_" + s + "_work.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s\\n\", odinmp_tacc_" + s + "_work.total);\n");
	    retval.append("      fprintf(stderr,\"  overhead:          time = %#12.7g s per thread\\n\", (odinmp_tacc_" + s + "_total.total - odinmp_tacc_" + s + "_work.total) / odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"                     time = %#12.7g s total\\n\", odinmp_tacc_" + s + "_total.total - odinmp_tacc_" + s + "_work.total);\n");
	    retval.append("      fprintf(stderr,\"                    prop. = %#12.7g %%\\n\", 100.0 * ((odinmp_tacc_" + s + "_total.total / odinmp_tacc_" + s + "_work.total) - 1.0));\n");
	    retval.append("   }\n");
	    retval.append("   fprintf(stderr,\"\\n\");\n");
	}
	return new String(retval);
    }

    public String declareTimingWsc(Enumeration e) {
	StringBuffer retval = new StringBuffer();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	}
	return new String(retval);
    }

    public String externalDeclareTimingWsc(Enumeration e) {
	StringBuffer retval = new StringBuffer();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	}
	return new String(retval);
    }

    public String initTimingWsc(Enumeration e) {
	StringBuffer retval = new StringBuffer();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_total);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_work);\n");
	}
	return new String(retval);
    }

    public String odinmpRoutines() {
	IncrementalReader ir;
	Enumeration e;

	try {
	    ir = new IncrementalReader(new BufferedReader(new FileReader(new File(supporting(), "odinmp.c"))));
	} catch (Exception ex) {
	    System.err.println("Caught Exception trying to open incremental file: ``" + ex + "''");
	    return null;
	}

	StringBuffer retval = new StringBuffer();
	retval.append(ir.readIncrement());
	retval.append("\n");
	retval.append("/* Common Type Definitions */\n");
	retval.append(commonTypedefs(new SymbolTable()));
	retval.append("\n");
	retval.append(PRINTF_define() + "\n");
	retval.append(TIMING_define() + "\n");
	retval.append("\n");
	retval.append("/* External Declarations */\n");
	retval.append("\n");
	retval.append(odinmpExternals());
	retval.append(ir.readIncrement());

	e = CriticalConstructDescriptor.names.elements();
	while (e.hasMoreElements()) {
	    retval.append("   pthread_mutex_init(&odinmp_crit_" + e.nextElement().toString() + ", ((void *)0));\n");
	}
	if (targetTimingMode() != mode_Never) {
	    Object o;

	    if (targetTimingMode() == mode_Runtime) {
		retval.append("   if (odinmp_timing) {\n");
	    }

	    retval.append(ParallelConstructDescriptor.initTiming());
	    retval.append(ForConstructDescriptor.initTiming());
	    retval.append(SectionsConstructDescriptor.initTiming());
	    retval.append(SingleConstructDescriptor.initTiming());
	    retval.append(BarrierConstructDescriptor.initTiming());

	    if (targetTimingMode() == mode_Runtime) {
		retval.append("   }\n");
	    }
	}
	

	retval.append(ir.readIncrement());
	if (targetTimingMode() != mode_Never) {
	    Object o;
	    String s;

	    if (targetTimingMode() == mode_Runtime) {
		retval.append("   if (odinmp_timing) {\n");
	    }

	    retval.append("    odinmp_timer_global_stopTime = odinmp_time();\n");
	    retval.append("    {\n");
	    retval.append("      double total = odinmp_timer_global_stopTime - odinmp_timer_global_preStartTime;\n");
	    retval.append("      double work = odinmp_timer_global_stopTime - odinmp_timer_global_startTime;\n");
	    retval.append("      fprintf(stderr,\"\\nProgram Execution:\\n\");\n");
	    retval.append("      fprintf(stderr,\"  total time:                   = %#12.7g s\\n\", total);\n");
	    retval.append("      fprintf(stderr,\"  working time:                 = %#12.7g s\\n\", work);\n");
	    retval.append("      fprintf(stderr,\"  odinmp startup overhead: time = %#12.7g s\\n\", total - work);\n");
	    retval.append("      fprintf(stderr,\"                        prop.   = %#12.7g %%\\n\", 100.0 * ((total / work) - 1.0));\n");
	    retval.append("      fprintf(stderr,\"\\n\");\n");
	    retval.append("    }\n");

	    retval.append(ParallelConstructDescriptor.reportTiming());
	    retval.append(ForConstructDescriptor.reportTiming());
	    retval.append(SectionsConstructDescriptor.reportTiming());
	    retval.append(SingleConstructDescriptor.reportTiming());
	    retval.append(BarrierConstructDescriptor.reportTiming());

	    retval.append("  fprintf(stderr,\"Waiting for threads to finish (odinmp_wait_finished()) (%d calls):\\n\", odinmp_wait_finished_tacc.n);\n");
	    retval.append("  fprintf(stderr,\"                                 min = %#12.7g s per call\\n\", odinmp_wait_finished_tacc.min);\n");
	    retval.append("  fprintf(stderr,\"                                 avg = %#12.7g s per call\\n\", odinmp_wait_finished_tacc.total / odinmp_wait_finished_tacc.n);\n");
	    retval.append("  fprintf(stderr,\"                                 max = %#12.7g s per call\\n\", odinmp_wait_finished_tacc.max);\n");
	    retval.append("  fprintf(stderr,\"                               total = %#12.7g s accumulated\\n\", odinmp_wait_finished_tacc.total);\n");
	    retval.append("  fprintf(stderr,\"\\n\");\n");

	    retval.append("  {\n");
	    retval.append("    int i;\n");
	    retval.append("    for (i = 0; i < odinmp_num_threads_allocated; i++) {\n");
	    retval.append("      fprintf(stderr,\"Thread %d spent a total of %#12.7g s waiting for mutex locks.\\n\", i, odinmp_thread_datas[i]->thread_info.timer_wait);\n");
	    retval.append("    }\n");
	    retval.append("  }\n");


	    if (targetTimingMode() == mode_Runtime) {
		retval.append("   }\n");
	    }
	}	

	retval.append(ir.readIncrement());
	retval.append("\n");
	

	e = CriticalConstructDescriptor.names.elements();
	while (e.hasMoreElements()) {
	    retval.append("pthread_mutex_t odinmp_crit_" + e.nextElement().toString() + ";\n");
	}

	if (targetTimingMode() != mode_Never) {
	    retval.append(ParallelConstructDescriptor.declareTiming());
	    retval.append(ForConstructDescriptor.declareTiming());
	    retval.append(SectionsConstructDescriptor.declareTiming());
	    retval.append(SingleConstructDescriptor.declareTiming());
	    retval.append(BarrierConstructDescriptor.declareTiming());
	}
	

	/* now create the global thread runner function */
	
	StringBuffer odinmp_thread = new StringBuffer();
	odinmp_thread.append("void odinmp_thread (odinmp_thread_data_t * odinmp_td)\n");
	odinmp_thread.append("{\n");
	odinmp_thread.append("  switch (" + OdinMP.THREAD_INFO("odinmp_td") + "->run_region) {\n");
	
	e = ParallelConstructDescriptor.inOrder.elements();
	while (e.hasMoreElements()) {
	    ParallelConstructDescriptor pcd = (ParallelConstructDescriptor)e.nextElement();
	    odinmp_thread.append("  case " + pcd.index + ":\n");
	    odinmp_thread.append("    odinmp_thread_" + pcd.sourceFile.mangledName + "(odinmp_td);\n");
	    odinmp_thread.append("    break;\n\n");
	}
	
	odinmp_thread.append("  default:\n");
	odinmp_thread.append("    fprintf(stderr, \"OdinMP: don't have a 'parallel' %d\\n\",\n");
	odinmp_thread.append("            " + OdinMP.THREAD_INFO("odinmp_td") + "->run_region)\n;");
	odinmp_thread.append("  }\n");
	odinmp_thread.append("}\n\n");
	
	/* now create the threadprivate initialization routine */
	
	StringBuffer odinmp_thread_init = new StringBuffer();
	odinmp_thread_init.append("void odinmp_thread_init (odinmp_thread_data_t *odinmp_td, int i) {\n");
	odinmp_thread_init.append("  switch (i) {\n");

	int i = 0;
	e = sourceFiles.elements();
	while (e.hasMoreElements()) {
	    SourceFile sf = (SourceFile)e.nextElement();
	    odinmp_thread_init.append("  case " + i + ":\n");
	    odinmp_thread_init.append("    " + OdinMP.PRINTF("\"thread %d: calling `odinmp_thread_init_" + sf.mangledName + "(odinmp_td, " + i + ")'\\n\", ((odinmp_thread_info_t *)odinmp_td)->id") + ";\n");
	    odinmp_thread_init.append("    odinmp_thread_init_" + sf.mangledName + "(odinmp_td, " + i + ");\n");
	    odinmp_thread_init.append("    break;\n");
	    i++;
	}
	odinmp_thread_init.append("  default:\n");
	    odinmp_thread_init.append("    " + OdinMP.PRINTF("\"thread %d: calling `odinmp_thread_spin'\\n\", ((odinmp_thread_info_t *)odinmp_td)->id") + ";\n");
	odinmp_thread_init.append("    odinmp_thread_spin(odinmp_td);\n");
	odinmp_thread_init.append("    break;\n");
	odinmp_thread_init.append("  }\n");
	odinmp_thread_init.append("}\n\n");

	try {
	    ir.close();
	} catch (Exception ex) {
	    System.err.println("Caught Exception trying to close incremental file: ``" + ex + "''");
	}
	
	retval.append(odinmp_thread);
	retval.append(odinmp_thread_init);
	
	return new String(retval);
    }

    public String odinmpExternals() {
	IncrementalReader ir;

	try {
	    ir = new IncrementalReader(new BufferedReader(new FileReader(new File(supporting(), "odinmp_external.h"))));
	} catch (Exception ex) {
	    System.err.println("Caught Exception trying to open incremental file: ``" + ex + "''");
	    return null;
	}

	StringBuffer retval = new StringBuffer();
	retval.append(ir.readIncrement());
	retval.append("\n");

	Enumeration e = CriticalConstructDescriptor.names.elements();
	while (e.hasMoreElements()) {
	    retval.append("extern pthread_mutex_t odinmp_crit_" + e.nextElement().toString() + ";\n");
	}
	if (targetTimingMode() != mode_Never) {
	    retval.append(ParallelConstructDescriptor.externalDeclareTiming());
	    retval.append(ForConstructDescriptor.externalDeclareTiming());
	    retval.append(SectionsConstructDescriptor.externalDeclareTiming());
	    retval.append(SingleConstructDescriptor.externalDeclareTiming());
	    retval.append(BarrierConstructDescriptor.externalDeclareTiming());
	}

	try {
	    ir.close();
	} catch (Exception ex) {
	    System.err.println("Caught Exception trying to close incremental file: ``" + ex + "''");
	}
	return  new String(retval);
    }

    public static void debugPrint(String s) {
	if (debugMode) {
	    System.err.print(s);
	}
    }

    public static void debugPrintln(String s) {
	/* debugPrint(s + "\n"); */
    }

    public static String PRINTF(String s) {
	switch (targetDebugMode()) {
	case mode_Always:
	    return "odinmp_printf(" + s + ")";
	case mode_Runtime:
	    return "if(odinmp_debug) odinmp_printf(" + s + ")";
	default:
	    return "";
	}
    }
    
    public static String PRINTF_define() {
	switch (targetDebugMode()) {
	case mode_Always:
	    return "#define PRINTF(x) odinmp_printf x";
	case mode_Runtime:
	    return "#define PRINTF(x) if(odinmp_debug) odinmp_printf x";
	default:
	    return "#define PRINTF(x) ";
	}
    }

    public static String TIMING_define() {
	switch (targetTimingMode()) {
	case mode_Always:
	    return 
		"#define ODINMP_TIMING\n" + 
		"#define ODINMP_TIMING_WAIT_BEGIN \\\n" +
		"{\\\n" +
		"  double odinmp_timer_wait_startTime = odinmp_time();\n" +
		"#define ODINMP_TIMING_WAIT_END \\\n" +
		"  ODINMP_THREAD_INFO(odinmp_self_data())->timer_wait += odinmp_time() - odinmp_timer_wait_startTime; \\\n" +
		"}\n";
	case mode_Runtime:
	    return 
		"#define ODINMP_TIMING\n" + 
		"#define ODINMP_TIMING_RUNTIME\n" + 
		"#define ODINMP_TIMING_WAIT_BEGIN \\\n" +
		"{\\\n" +
		"  double odinmp_timer_wait_startTime;\\\n" +
		"  if(odinmp_timing) \\\n" +
		"    odinmp_timer_wait_startTime = odinmp_time();\n" +
		"#define ODINMP_TIMING_WAIT_END \\\n" +
		"  if (odinmp_timing) \\\n" +
		"    ODINMP_THREAD_INFO(odinmp_self_data())->timer_wait += odinmp_time() - odinmp_timer_wait_startTime; \\\n" +
		"}\n";
	default:
	    return 
		"#define ODINMP_TIMING_WAIT_BEGIN\n" +
		"#define ODINMP_TIMING_WAIT_END\n";
	}
    }

    public static String THREAD_INFO(String s) {
	return "((odinmp_thread_info_t *)(" + s + "))";
    }

    public void dsc(String s) {
    }
}

