/************************************************************
 * program to solve a finite difference 
 * discretization of Helmholtz equation :  
 * (d2/dx2)u + (d2/dy2)u - alpha u = f 
 * using Jacobi iterative method. 
 *
 * Modified: Sanjiv Shah,       Kuck and Associates, Inc. (KAI), 1998
 * Author:   Joseph Robicheaux, Kuck and Associates, Inc. (KAI), 1998
 * 
 * Directives are used in this code to achieve paralleism. 
 * All do loops are parallized with default 'static' scheduling.
 * 
 * Input :  n - grid dimension in x direction 
 *          m - grid dimension in y direction
 *          alpha - Helmholtz constant (always greater than 0.0)
 *          tol   - error tolerance for iterative solver
 *          relax - Successice over relaxation parameter
 *          mits  - Maximum iterations for iterative solver
 *
 * On output 
 *       : u(n,m) - Dependent variable (solutions)
 *       : f(n,m) - Right hand side function 
 *************************************************************/

#include <math.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>

int n,m,mits,mtemp;
double tol,relax,alpha;

void error_check (int n, int m, double alpha,
                  double dx, double dy,
                  double **u, double **f) 
{
/************************************************************
 * Checks error between numerical and exact solution 
 *
 ************************************************************/
     
      int i,j;
      double xx,yy,temp,error;

      dx = 2.0 / (n-1);
      dy = 2.0 / (m-1);
      error = 0.0;

      #pragma omp parallel for private(xx,yy,temp) reduction(+:error)
      for (i = 0; i < n; i++) {
         for (j = 0; j < m; j++) {
            xx = -1.0 + dx * (double)i;
            yy = -1.0 + dy * (double)j;
            temp  = u[i][j] - (1.0-xx*xx)*(1.0-yy*yy);
            error = error + temp*temp;
         }
      }
  
      error = sqrt(error)/(double)(n*m);

      printf("Solution Error : %e\n",error);

      return;
}


void jacobi (int n, int m, double dx, double dy, 
             double alpha, double omega,
             double **u, double **f,
             double tol, int maxit)
{

/******************************************************************
 * Subroutine HelmholtzJ
 * Solves poisson equation on rectangular grid assuming : 
 * (1) Uniform discretization in each direction, and 
 * (2) Dirichlect boundary conditions 
 * 
 * Jacobi method is used in this routine 
 *
 * Input : n,m   Number of grid points in the X/Y directions 
 *         dx,dy Grid spacing in the X/Y directions 
 *         alpha Helmholtz eqn. coefficient 
 *         omega Relaxation factor 
 *         f(n,m) Right hand side function 
 *         u(n,m) Dependent variable/Solution
 *         tol    Tolerance for iterative solver 
 *         maxit  Maximum number of iterations 
 *
 * Output : u(n,m) - Solution 
 *****************************************************************/

/*
 * Local variables 
 */
      int i,j,k,k_local;
      double error,resid,rsum,ax,ay,b;
      double error_local, **uold;

      float ta,tb,tc,td,te,ta1,ta2,tb1,tb2,tc1,tc2,td1,td2;
      float te1,te2;
      float second;

      uold = (double **)malloc(n * sizeof(double *));  
      for (i = 0; i < n; i++)
          uold[i] = (double *)malloc(m * sizeof(double));

/*
 * Initialize coefficients 
 */

      ax = 1.0/(dx*dx); /* X-direction coef */
      ay = 1.0/(dy*dy); /* Y-direction coef*/
      b  = -2.0/(dx*dx)-2.0/(dy*dy) - alpha; /* Central coeff  */

      error = 10.0 * tol; 
      k = 1;

      while (k < maxit && error > tol) {

         error = 0.0;

/* Copy new solution into old */

	 #pragma omp parallel
	 {
	   
	   #pragma omp for private(i, j)
	   for (i = 0; i < n; i++) {
	     for (j = 0; j < m; j++) {
               uold[i][j] = u[i][j]; 
	     }
	   } 
	   
/* Compute stencil, residual, & update*/
	   
	   #pragma omp for private(i, j, resid) reduction(+:error) nowait
	   for (i = 1; i < n-1; i++) {
	     for (j = 1; j < m-1; j++) {
/*     Evaluate residual */
               resid = (ax*(uold[i-1][j] + uold[i+1][j]) 
			+ ay*(uold[i][j-1] + uold[i][j+1])
			+ b * uold[i][j] - f[i][j])/b;
/* Update solution */
               u[i][j] = uold[i][j] - omega * resid;
/* Accumulate residual error*/
               error = error + resid*resid;
	     }
	   }
	 } /* #pragma omp parallel */

/* Error check */

         k = k + 1;
	 
         error = sqrt(error)/(double)(n*m);

      }                     /* End iteration loop */

      for (i = 0; i < n; i++)
          free(uold[i]);
      free(uold);

      printf ("Total Number of Iterations: %d\n", k);
      printf ("Residual error            : %e\n", error);

      return;
}

void initialize (int n, int m, double alpha,
                 double *dx_p, double *dy_p, 
                 double **u, double **f)
{

/******************************************************
 * Initializes data 
 * Assumes exact solution is u(x,y) = (1-x^2)*(1-y^2)
 *
 ******************************************************/
     
      int i,j;
      double xx, yy, dx, dy;

      dx = 2.0 / (n-1);
      dy = 2.0 / (m-1);

/* Initilize initial condition and RHS*/

      #pragma omp parallel for private(i,j,xx,yy)
      for (i = 0; i < n; i++) {
	double *ul = u[i];
	double *fl = f[i];
	xx = (dx * i) - 1.0;          /* -1 < x < 1*/
	for (j = 0; j < m; j++) {
	  yy = (dy * j) - 1.0;        /* -1 < y < 1*/
	  ul[j] = 0.0; 
	  fl[j] = -alpha * (1.0-xx*xx) * (1.0-yy*yy)
	    - 2.0 * (1.0-xx*xx) - 2.0 * (1.0-yy*yy);
	}
      }

      *dx_p = dx;
      *dy_p = dy;

      return;
}

void driver () 
{

/*************************************************************
 * Subroutine driver () 
 * This is where the arrays are allocated and initialzed. 
 *
 * Working varaibles/arrays 
 *     dx  - grid spacing in x direction 
 *     dy  - grid spacing in y direction 
 *************************************************************/

      double **u, **f, dx, dy;
      int i, j;

      u = (double **)malloc(n * sizeof(double *));
      for (i = 0; i < n; i++)
          u[i] = (double *)malloc(m * sizeof(double));

      f = (double **)malloc(n * sizeof(double *));
      for (i = 0; i < n; i++)
          f[i] = (double *)malloc(m * sizeof(double));

/* Initialize data */

      initialize (n,m,alpha,&dx,&dy,u,f);

/* Solve Helmholtz equation */

      jacobi (n,m,dx,dy,alpha,relax,u,f,tol,mits);


/* Check error between exact solution */

      error_check (n,m,alpha,dx,dy,u,f);

      for (i = 0; i < n; i++)
          free(u[i]);
      free(u);

      for (i = 0; i < n; i++)
          free(f[i]);
      free(f);

      return; 
}


int main(int argc, char **argv) {

  struct timeval t0, t1;
  gettimeofday(&t0,NULL);

/*
 * Read info 
 */ 
      printf("Input n,m - grid dimension in x,y direction: "); 
      scanf("%d %d", &n, &m);
      printf("n = %d, m = %d\n", n, m);
      printf("Input alpha - Helmholts constant: ");
      scanf("%lf", &alpha);
      printf("alpha = %e\n", alpha);
      printf("Input relax - Successive over-relaxation parameter: ");
      scanf("%lf", &relax);
      printf("relax = %e\n", relax);
      printf("Input tol - error tolerance for iterative solver: ");
      scanf("%lf", &tol);
      printf("tol = %e\n", tol);
      printf("Input mits - Maximum iterations for solver: ");
      scanf("%d", &mits);
      printf("mits = %d\n", mits);

/*
 * Calls a driver routine 
 */ 
      driver();

      gettimeofday(&t1,NULL);
      fprintf(stderr,"Program execution time: %.7g s\n",
	      t1.tv_sec - t0.tv_sec + (t1.tv_usec - t0.tv_usec)/1000000.0);
      
      exit(0);
}


